package com.jayway.brain

import collection.JavaConversions._
import collection._
import se.citerus.crazysnake.{Position, HeatState}
import com.jayway.brain.Utils._


case class ClusterSpace(state: HeatState, myHeadPosition: Position, opponentHeadPosition: Position) {

  /**
   *             x_range    2x_range
   *           *********************
   *           *         |         *
   *  y_range  *    A    |     B   *
   *           *         |         *
   *           *---------|---------*
   *           *         |         *
   * 2y_range  *    C    |     D   *
   *           *         |         *
   *           *********************
   */

  val y_range = state.getNoRows / 2
  val x_range = state.getNoColumns / 2

  val boundary = 10

  val fruits = state.getFruitPositions
  val section_A: mutable.ListBuffer[Position] = mutable.ListBuffer[Position]()
  val section_B: mutable.ListBuffer[Position] = mutable.ListBuffer[Position]()
  val section_C: mutable.ListBuffer[Position] = mutable.ListBuffer[Position]()
  val section_D: mutable.ListBuffer[Position] = mutable.ListBuffer[Position]()

  fruits.foreach {
    f => f match {
      case p if (p.getY >= 0 && p.getY < y_range && p.getX >= 0 && p.getX < x_range) => section_A += p
      case p if (p.getY >= 0 && p.getY < y_range && p.getX >= x_range && p.getX < 2 * x_range) => section_B += p
      case p if (p.getY >= y_range && p.getY < 2 * y_range && p.getX >= 0 && p.getX < x_range) => section_C += p
      case p if (p.getY >= y_range && p.getY < 2 * y_range && p.getX >= x_range && p.getX < 2 * x_range) => section_D += p
    }
  }

  val A: mutable.ListBuffer[(Int, Int, Int, Position)] = section_A.map {
    p => (section_A.size, calculateH(myHeadPosition, p), calculateH(opponentHeadPosition, p), p)
  }
  val B: mutable.ListBuffer[(Int, Int, Int, Position)] = section_B.map {
    p => (section_B.size, calculateH(myHeadPosition, p), calculateH(opponentHeadPosition, p), p)
  }
  val C: mutable.ListBuffer[(Int, Int, Int, Position)] = section_C.map {
    p => (section_C.size, calculateH(myHeadPosition, p), calculateH(opponentHeadPosition, p), p)
  }
  val D: mutable.ListBuffer[(Int, Int, Int, Position)] = section_D.map {
    p => (section_D.size, calculateH(myHeadPosition, p), calculateH(opponentHeadPosition, p), p)
  }

  val result = (A ++ B ++ C ++ D).sortWith {
    (p1, p2) =>
      if (p1._1.compareTo(p2._1) == 0)
        if (p1._2.compareTo(p2._2) == 0)
          p1._3.compareTo(p2._3) > 0
        else
          p1._2.compareTo(p2._2) < 0
      else
        p1._1.compareTo(p2._1) > 0
  }

  val closeFruit = result.filter {
    f => f._2 <= boundary
  }

  def getBestFruit: Position = {
    if (closeFruit.isEmpty)
      result.head._4
    else
      closeFruit.head._4
  }
}
