package com.jayway.brain

import se.citerus.crazysnake._
import collection.JavaConversions._
import collection._
import com.jayway.brain.Utils._

class Snakeway extends BaseBrain {

  var state: HeatState = _

  var openList: mutable.Set[Node] = mutable.Set[Node]()
  var closedList: mutable.Set[Node] = mutable.Set[Node]()

  var currentFruit: Position = null

  override def getNextMove(state: HeatState) = {

    val level = 20

    this.state = state

    val me = getMe(state)
    val opponent = getOpponent(state)
    val myHeadPosition = me.getHeadPosition
    val opponentHeadPosition = opponent.getHeadPosition

    selectFruit(myHeadPosition, opponentHeadPosition)

    var nextPosition = path(runAstar(Node(myHeadPosition), currentFruit, level), currentFruit).head.position
    nextPosition = avoidOpponent(nextPosition, myHeadPosition, opponentHeadPosition)
    interpretNextMove(myHeadPosition, nextPosition, me.getDirection)
  }

  def selectFruit(myHeadPosition: Position, opponentHeadPosition: Position) {
    if (state.getFruitPositions.size() == 0) {
      val r = new scala.util.Random
      currentFruit = new Position(r.nextInt(state.getNoColumns), r.nextInt(state.getNoRows))
    } else
      currentFruit = ClusterSpace(state, myHeadPosition, opponentHeadPosition).getBestFruit
  }

  override def getName = "Snakeway"

  def interpretNextMove(headPosition: Position, nextPosition: Position, snakeDirection: Direction) =
    (snakeDirection, nextPosition.getX - headPosition.getX, nextPosition.getY - headPosition.getY) match {
      case (Direction.NORTH, 0, 1) => Movement.LEFT // this case only happens if snake length is 1
      case (Direction.NORTH, -1, 0) => Movement.LEFT
      case (Direction.NORTH, 1, 0) => Movement.RIGHT
      case (Direction.NORTH, 0, -1) => Movement.FORWARD
      case (Direction.SOUTH, 0, -1) => Movement.LEFT // this case only happens if snake length is 1
      case (Direction.SOUTH, -1, 0) => Movement.RIGHT
      case (Direction.SOUTH, 1, 0) => Movement.LEFT
      case (Direction.SOUTH, 0, 1) => Movement.FORWARD
      case (Direction.EAST, 0, 1) => Movement.RIGHT
      case (Direction.EAST, 0, -1) => Movement.LEFT
      case (Direction.EAST, 1, 0) => Movement.FORWARD
      case (Direction.EAST, -1, 0) => Movement.RIGHT // this case only happens if snake length is 1
      case (Direction.WEST, 0, 1) => Movement.LEFT
      case (Direction.WEST, -1, 0) => Movement.FORWARD
      case (Direction.WEST, 0, -1) => Movement.RIGHT
      case (Direction.WEST, 1, 0) => Movement.LEFT // this case only happens if snake length is 1
    }

  def avoidOpponent(decidedPosition: Position, myHead: Position, opponentHead: Position): Position = {
    neighborOf(decidedPosition).contains(opponentHead) match {
      case true => availableNeighborOf(myHead, decidedPosition).take(1).headOption match {
        case Some(x) => x
        case None => decidedPosition
      }
      case false => decidedPosition
    }
  }

  def neighborOf(n: Position): mutable.Set[Position] = {
    val x = n.getX
    val y = n.getY
    var nodes = mutable.Set[Position]()
    nodes += new Position(x + 1, y)
    nodes += new Position(x - 1, y)
    nodes += new Position(x - 1, y + 1)
    nodes += new Position(x, y + 1)
    nodes += new Position(x + 1, y + 1)
    nodes += new Position(x - 1, y - 1)
    nodes += new Position(x, y - 1)
    nodes += new Position(x + 1, y - 1)
    nodes
  }

  def availableNeighborOf(node: Node, position: Position): mutable.Set[Node] = {
    val x = node.position.getX
    val y = node.position.getY
    var nodes = mutable.Set[Node]()
    if (x < state.getNoColumns) {
      val n = Node(new Position(x + 1, y))
      n.g = node.g + 1
      n.h = calculateH(n.position, position)
      n.parent = node
      nodes += n
    }
    if (y < state.getNoColumns) {
      val n = Node(new Position(x, y + 1))
      n.g = node.g + 1
      n.h = calculateH(n.position, position)
      n.parent = node
      nodes += n
    }
    if (x > 0) {
      val n = Node(new Position(x - 1, y))
      n.g = node.g + 1
      n.h = calculateH(n.position, position)
      n.parent = node
      nodes += n
    }
    if (y > 0) {
      val n = Node(new Position(x, y - 1))
      n.g = node.g + 1
      n.h = calculateH(n.position, position)
      n.parent = node
      nodes += n
    }
    nodes filter {
      x => !state.getSquare(x.position).containsSnake() && !state.getSquare(x.position).containsWall() && !state.getSquare(x.position).hasSolidContent
    }
  }

  // Overloaded function
  def availableNeighborOf(position: Position, conflictPosition: Position): mutable.Set[Position] = {
    val x = position.getX
    val y = position.getY
    var nodes = mutable.Set[Position]()
    if (x < state.getNoColumns) nodes += new Position(x + 1, y)
    if (y < state.getNoColumns) nodes += new Position(x, y + 1)
    if (x > 0) nodes += new Position(x - 1, y)
    if (y > 0) nodes += new Position(x, y - 1)
    nodes filter {
      x => !state.getSquare(x).containsSnake() && !state.getSquare(x).containsWall() && !state.getSquare(x).hasSolidContent && x != conflictPosition
    }
  }

  def runAstar(node: Node, position: Position, level: Int): Node = {
    closedList.clear()
    openList.clear()
    node.h = calculateH(node.position, position)
    openList += node
    var bestNode: Node = null
    while (!openList.isEmpty) {
      bestNode = bestFinOpenList()
      if (bestNode.position == position)
        return bestNode
      closedList += bestNode
      openList -= bestNode
      val neighbors = availableNeighborOf(bestNode, position)
      var g = Integer.MAX_VALUE
      neighbors.foreach {
        n =>
          if ((closedList.contains(n) || openList.contains(n)) && n.g <= g) {
            g = n.g
            n.parent = bestNode
            if (g >= level) return bestNode
          } else
            openList += n
      }
    }
    bestNode
  }

  def path(node: Node, fruit: Position): List[Node] = {
    def getPath(node: Node): List[Node] = node.parent match {
      case null => List()
      case p: Node => getPath(p) ::: List(p)
    }
    getPath(node).drop(1) ::: List(node)
  }

  def bestFinOpenList(): Node = {
    openList.min
  }

}

case class Node(position: Position) extends Ordered[Node] {
  var g: Int = 0
  var h: Int = 0
  var parent: Node = _

  def f: Int = g + h

  override def compare(otherNode: Node): Int = f - otherNode.f

}